/*
 * linux/arch/sh/kernel/setup_hs7751rvoip.c
 *
 * Copyright (C) 2000  Kazumoto Kojima
 *
 * Renesas Technology Sales HS7751RVoIP Support.
 *
 * Modified for HS7751RVoIP by
 * Atom Create Engineering Co., Ltd. 2002.
 */

#include <linux/config.h>
#include <linux/init.h>
#include <linux/irq.h>

#include <linux/hdreg.h>
#include <linux/ide.h>
#include <linux/pci.h>
#include <asm/io.h>
#include <asm/hitachi_hs7751rvoip.h>

#include <linux/mm.h>
#include <linux/vmalloc.h>

static int mask_pos[] = {8, 9, 10, 11, 12, 13, 0, 1, 2, 3, 4, 5, 6, 7};

/* defined in mm/ioremap.c */
extern void * p3_ioremap(unsigned long phys_addr, unsigned long size, unsigned long flags);

static unsigned int debug_counter;

static void enable_hs7751rvoip_irq(unsigned int irq);
static void disable_hs7751rvoip_irq(unsigned int irq);

/* shutdown is same as "disable" */
#define shutdown_hs7751rvoip_irq disable_hs7751rvoip_irq

static void ack_hs7751rvoip_irq(unsigned int irq);
static void end_hs7751rvoip_irq(unsigned int irq);

static unsigned int startup_hs7751rvoip_irq(unsigned int irq)
{
	enable_hs7751rvoip_irq(irq);
	return 0; /* never anything pending */
}

static void disable_hs7751rvoip_irq(unsigned int irq)
{
	unsigned long flags;
	unsigned short val;
	unsigned short mask = 0xffff ^ (0x0001 << mask_pos[irq]);

	/* Set the priority in IPR to 0 */
	save_and_cli(flags);
	val = ctrl_inw(IRLCNTR3);
	val &= mask;
	ctrl_outw(val, IRLCNTR3);
	restore_flags(flags);
}

static void enable_hs7751rvoip_irq(unsigned int irq)
{
	unsigned long flags;
	unsigned short val;
	unsigned short value = (0x0001 << mask_pos[irq]);

	/* Set priority in IPR back to original value */
	save_and_cli(flags);
	val = ctrl_inw(IRLCNTR3);
	val |= value;
	ctrl_outw(val, IRLCNTR3);
	restore_flags(flags);
}

static void ack_hs7751rvoip_irq(unsigned int irq)
{
	disable_hs7751rvoip_irq(irq);
}

static void end_hs7751rvoip_irq(unsigned int irq)
{
	if (!(irq_desc[irq].status & (IRQ_DISABLED|IRQ_INPROGRESS)))
		enable_hs7751rvoip_irq(irq);
}

static struct hw_interrupt_type hs7751rvoip_irq_type = {
	"HS7751RVoIP IRQ",
	startup_hs7751rvoip_irq,
	shutdown_hs7751rvoip_irq,
	enable_hs7751rvoip_irq,
	disable_hs7751rvoip_irq,
	ack_hs7751rvoip_irq,
	end_hs7751rvoip_irq,
};

static void make_hs7751rvoip_irq(unsigned int irq)
{
	disable_irq_nosync(irq);
	irq_desc[irq].handler = &hs7751rvoip_irq_type;
	disable_hs7751rvoip_irq(irq);
}

/*
 * Initialize IRQ setting
 */
void __init init_hs7751rvoip_IRQ(void)
{
	int i;

	/* IRL0=ON HOOK1
	 * IRL1=OFF HOOK1
	 * IRL2=ON HOOK2
	 * IRL3=OFF HOOK2
	 * IRL4=Ringing Detection
	 * IRL5=CODEC
	 * IRL6=Ethernet
	 * IRL7=Ethernet Hub
	 * IRL8=USB Communication
	 * IRL9=USB Connection
	 * IRL10=USB DMA
	 * IRL11=CF Card
	 * IRL12=PCMCIA
	 * IRL13=PCI Slot
	 */
	ctrl_outw(0x9876, IRLCNTR1);
	ctrl_outw(0xdcba, IRLCNTR2);
	ctrl_outw(0x0050, IRLCNTR4);
	ctrl_outw(0x4321, IRLCNTR5);

	for (i=0; i<14; i++)
		make_hs7751rvoip_irq(i);
}

/*
 * Initialize the board
 */
void __init setup_hs7751rvoip(void)
{
#if defined(CONFIG_HS7751RVOIP2)
	printk(KERN_INFO "Renesas Technology Sales HS7751RVoIP-2 support.\n");
#else
	printk(KERN_INFO "Hitachi Semiconductor and Devices HS7751RVoIP-1 support.\n");
#endif
	ctrl_outb(0xf0, PA_OUTPORTR);
	debug_counter = 0;
}

void *area5_io8_base;
void *area6_io8_base;
void *area5_io16_base;
void *area6_io16_base;

int __init cf_init(void)
{
	pgprot_t prot;
	unsigned long paddrbase, psize;

	/* open I/O area window */
#if defined(CONFIG_HS7751RVOIP2)
	paddrbase = virt_to_phys((void *)(PA_AREA5_IO+0x00000800));
#else
	paddrbase = virt_to_phys((void *)PA_AREA5_IO);
#endif
	psize = PAGE_SIZE;
#if defined(CONFIG_HS7751RVOIP2)
	prot = PAGE_KERNEL_PCC(1, _PAGE_PCC_COM16);
#else
	prot = PAGE_KERNEL_PCC(1, _PAGE_PCC_IO16);
#endif
	area5_io16_base = p3_ioremap(paddrbase, psize, prot.pgprot);
	if (!area5_io16_base) {
		printk("allocate_cf_area : can't open CF I/O 16bit window!\n");
		return -ENOMEM;
	}
#if !defined(CONFIG_HS7751RVOIP2)
#if defined(CONFIG_HS7751RVOIP2)
	prot = PAGE_KERNEL_PCC(1, _PAGE_PCC_COM8);
#else
	prot = PAGE_KERNEL_PCC(1, _PAGE_PCC_IO8);
#endif
	area5_io8_base = p3_ioremap(paddrbase, psize, prot.pgprot);
	if (!area5_io8_base) {
		printk("allocate_cf_area : can't open CF I/O 8bit window!\n");
		return -ENOMEM;
	}
#endif

	/* XXX : do we need attribute and common-memory area also? */

	paddrbase = virt_to_phys((void *)PA_AREA6_IO);
	psize = PAGE_SIZE;
#if defined(CONFIG_HS7751RVOIP_CODEC)
	prot = PAGE_KERNEL_PCC(0, _PAGE_PCC_COM8);
#else
	prot = PAGE_KERNEL_PCC(0, _PAGE_PCC_IO8);
#endif
	area6_io8_base = p3_ioremap(paddrbase, psize, prot.pgprot);
	if (!area6_io8_base) {
		printk("allocate_cf_area : can't open CODEC I/O 8bit window!\n");
		return -ENOMEM;
	}
	prot = PAGE_KERNEL_PCC(0, _PAGE_PCC_IO16);
	area6_io16_base = p3_ioremap(paddrbase, psize, prot.pgprot);
	if (!area6_io16_base) {
		printk("allocate_cf_area : can't open CODEC I/O 16bit window!\n");
		return -ENOMEM;
	}

	return 0;
}

__initcall (cf_init);

void debug_led_disp(void)
{
	unsigned char value;

	value = (unsigned char)debug_counter++;
	ctrl_outb((0xf0|value), PA_OUTPORTR);
	if (value == 0x0f)
		debug_counter = 0;
}
