/*
 * linux/ipc/util.h
 * Copyright (C) 1999 Christoph Rohland
 *
 * ipc helper functions (c) 1999 Manfred Spraul <manfreds@colorfullife.com>
 */

#define USHRT_MAX 0xffff
#define SEQ_MULTIPLIER	(IPCMNI)

void sem_init (void);
void msg_init (void);
void shm_init (void);

struct ipc_ids {
	int size;
	int in_use;
	int max_id;
	unsigned short seq;
	unsigned short seq_max;
	struct semaphore sem;	
	spinlock_t ary;
	struct ipc_id* entries;
};

struct ipc_id {
	struct kern_ipc_perm* p;
};


void __init ipc_init_ids(struct ipc_ids* ids, int size);

/* must be called with ids->sem acquired.*/
int ipc_findkey(struct ipc_ids* ids, key_t key);
int ipc_addid(struct ipc_ids* ids, struct kern_ipc_perm* new, int size);

/* must be called with both locks acquired. */
struct kern_ipc_perm* ipc_rmid(struct ipc_ids* ids, int id);

int ipcperms (struct kern_ipc_perm *ipcp, short flg);

/* for rare, potentially huge allocations.
 * both function can sleep
 */
void* ipc_alloc(int size);
void ipc_free(void* ptr, int size);

#if defined(CONFIG_OPTIMIZE_ZERO)
static inline void ipc_lockall(struct ipc_ids* ids)
#else
extern inline void ipc_lockall(struct ipc_ids* ids)
#endif
{
	spin_lock(&ids->ary);
}

#if defined(CONFIG_OPTIMIZE_ZERO)
static inline struct kern_ipc_perm* ipc_get(struct ipc_ids* ids, int id)
#else
extern inline struct kern_ipc_perm* ipc_get(struct ipc_ids* ids, int id)
#endif
{
	struct kern_ipc_perm* out;
	int lid = id % SEQ_MULTIPLIER;
	if(lid >= ids->size)
		return NULL;

	out = ids->entries[lid].p;
	return out;
}

#if defined(CONFIG_OPTIMIZE_ZERO)
static inline void ipc_unlockall(struct ipc_ids* ids)
#else
extern inline void ipc_unlockall(struct ipc_ids* ids)
#endif
{
	spin_unlock(&ids->ary);
}
#if defined(CONFIG_OPTIMIZE_ZERO)
static inline struct kern_ipc_perm* ipc_lock(struct ipc_ids* ids, int id)
#else
extern inline struct kern_ipc_perm* ipc_lock(struct ipc_ids* ids, int id)
#endif
{
	struct kern_ipc_perm* out;
	int lid = id % SEQ_MULTIPLIER;
	if(lid >= ids->size)
		return NULL;

	spin_lock(&ids->ary);
	out = ids->entries[lid].p;
	if(out==NULL)
		spin_unlock(&ids->ary);
	return out;
}

#if defined(CONFIG_OPTIMIZE_ZERO)
static inline void ipc_unlock(struct ipc_ids* ids, int id)
#else
extern inline void ipc_unlock(struct ipc_ids* ids, int id)
#endif
{
	spin_unlock(&ids->ary);
}

#if defined(CONFIG_OPTIMIZE_ZERO)
static inline int ipc_buildid(struct ipc_ids* ids, int id, int seq)
#else
extern inline int ipc_buildid(struct ipc_ids* ids, int id, int seq)
#endif
{
	return SEQ_MULTIPLIER*seq + id;
}

#if defined(CONFIG_OPTIMIZE_ZERO)
static inline int ipc_checkid(struct ipc_ids* ids, struct kern_ipc_perm* ipcp, int uid)
#else
extern inline int ipc_checkid(struct ipc_ids* ids, struct kern_ipc_perm* ipcp, int uid)
#endif
{
	if(uid/SEQ_MULTIPLIER != ipcp->seq)
		return 1;
	return 0;
}

void kernel_to_ipc64_perm(struct kern_ipc_perm *in, struct ipc64_perm *out);
void ipc64_perm_to_ipc_perm(struct ipc64_perm *in, struct ipc_perm *out);

#ifdef __ia64__
  /* On IA-64, we always use the "64-bit version" of the IPC structures.  */ 
# define ipc_parse_version(cmd)	IPC_64
#else
int ipc_parse_version (int *cmd);
#endif
